from typing import Callable
import numpy as np
import pytest

from neuraxle.base import Identity
from neuraxle.hyperparams.distributions import Boolean
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.pipeline import Pipeline
from neuraxle.steps.flow import ChooseOneOrManyStepsOf, ChooseOneStepOf
from neuraxle.steps.misc import TransformCallbackStep, TapeCallbackFunction, FitTransformCallbackStep, AssertFalseStep
from testing.steps.neuraxle_test_case import NeuraxleTestCase

DATA_INPUTS = np.array(range(10))
EXPECTED_OUTPUTS = np.array(range(10))


def create_test_case_single_step_choosen():
    a_callback = TapeCallbackFunction()
    b_callback = TapeCallbackFunction()

    return NeuraxleTestCase(
        pipeline=Pipeline([
            ChooseOneOrManyStepsOf([
                ('a', TransformCallbackStep(a_callback, transform_function=lambda di: di * 2)),
                ('b', TransformCallbackStep(b_callback, transform_function=lambda di: di * 2))
            ]),
        ]),
        callbacks=[a_callback, b_callback],
        expected_callbacks_data=[
            DATA_INPUTS,
            []
        ],
        hyperparams={
            'ChooseOneOrManyStepsOf__a__enabled': True,
            'ChooseOneOrManyStepsOf__b__enabled': False
        },
        hyperparams_space={
            'ChooseOneOrManyStepsOf__a__enabled': Boolean(),
            'ChooseOneOrManyStepsOf__b__enabled': Boolean()
        },
        expected_processed_outputs=np.array([0, 2, 4, 6, 8, 10, 12, 14, 16, 18])
    )


def create_test_case_multiple_steps_choosen():
    a_callback = TapeCallbackFunction()
    b_callback = TapeCallbackFunction()

    return NeuraxleTestCase(
        pipeline=Pipeline([
            ChooseOneOrManyStepsOf([
                ('a', TransformCallbackStep(a_callback, transform_function=lambda di: di * 2)),
                ('b', TransformCallbackStep(b_callback, transform_function=lambda di: di * 2))
            ]),
        ]),
        callbacks=[a_callback, b_callback],
        expected_callbacks_data=[DATA_INPUTS, DATA_INPUTS],
        hyperparams={
            'ChooseOneOrManyStepsOf__a__enabled': True,
            'ChooseOneOrManyStepsOf__b__enabled': True
        },
        hyperparams_space={
            'ChooseOneOrManyStepsOf__a__enabled': Boolean(),
            'ChooseOneOrManyStepsOf__b__enabled': Boolean()
        },
        expected_processed_outputs=np.array([0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 0, 2, 4, 6, 8, 10, 12, 14, 16, 18])
    )


def create_test_case_invalid_step_choosen():
    a_callback = TapeCallbackFunction()
    b_callback = TapeCallbackFunction()

    return NeuraxleTestCase(
        pipeline=Pipeline([
            ChooseOneOrManyStepsOf([
                ('a', TransformCallbackStep(a_callback, transform_function=lambda di: di * 2)),
                ('b', TransformCallbackStep(b_callback, transform_function=lambda di: di * 2))
            ]),
        ]),
        callbacks=[a_callback, b_callback],
        expected_callbacks_data=[DATA_INPUTS, DATA_INPUTS],
        hyperparams={
            'ChooseOneOrManyStepsOf__c__enabled': True,
            'ChooseOneOrManyStepsOf__b__enabled': False
        },
        hyperparams_space={
            'ChooseOneOrManyStepsOf__a__enabled': Boolean(),
            'ChooseOneOrManyStepsOf__b__enabled': Boolean()
        },
        expected_processed_outputs=np.array([0, 2, 4, 6, 8, 10, 12, 14, 16, 18])
    )


def create_test_case_invalid_step_not_choosen():
    a_callback = TapeCallbackFunction()
    b_callback = TapeCallbackFunction()

    return NeuraxleTestCase(
        pipeline=Pipeline([
            ChooseOneOrManyStepsOf([
                ('a', TransformCallbackStep(a_callback, transform_function=lambda di: di * 2)),
                ('b', TransformCallbackStep(b_callback, transform_function=lambda di: di * 2))
            ]),
        ]),
        callbacks=[a_callback, b_callback],
        expected_callbacks_data=[DATA_INPUTS, DATA_INPUTS],
        hyperparams={
            'ChooseOneOrManyStepsOf__c__enabled': False,
            'ChooseOneOrManyStepsOf__b__enabled': False
        },
        hyperparams_space={
            'ChooseOneOrManyStepsOf__a__enabled': Boolean(),
            'ChooseOneOrManyStepsOf__b__enabled': Boolean()
        },
        expected_processed_outputs=np.array(range(10))
    )


@pytest.mark.parametrize('test_case', [
    create_test_case_single_step_choosen(),
    create_test_case_multiple_steps_choosen()
])
def test_choose_one_or_many_step_of_transform_should_choose_step(
        test_case: NeuraxleTestCase):
    p = test_case.pipeline
    test = HyperparameterSpace(test_case.hyperparams_space)
    p.set_hyperparams_space(test)
    p.set_hyperparams(test_case.hyperparams)

    outputs = p.transform(DATA_INPUTS)

    assert np.array_equal(outputs, test_case.expected_processed_outputs)
    assert_callback_data_is_as_expected(test_case)


def assert_callback_data_is_as_expected(test_case):
    for callback, expected_callback_data in zip(test_case.callbacks, test_case.expected_callbacks_data):
        if len(callback.data) > 0:
            assert np.array_equal(
                np.array(callback.data[0]),
                expected_callback_data
            )
        else:
            assert np.array_equal(
                np.array([]),
                np.array(expected_callback_data)
            )


def create_test_case_fit_transform_single_step_choosen():
    a_callback = TapeCallbackFunction()
    b_callback = TapeCallbackFunction()
    c_callback = TapeCallbackFunction()
    d_callback = TapeCallbackFunction()

    return NeuraxleTestCase(
        pipeline=Pipeline([
            ChooseOneOrManyStepsOf([
                ('a', FitTransformCallbackStep(a_callback, c_callback, transform_function=lambda di: di * 2)),
                ('b', FitTransformCallbackStep(b_callback, d_callback, transform_function=lambda di: di * 2))
            ]),
        ]),
        callbacks=[a_callback, c_callback, b_callback, d_callback],
        expected_callbacks_data=[
            DATA_INPUTS,
            (DATA_INPUTS, EXPECTED_OUTPUTS),
            [],
            []
        ],
        hyperparams={
            'ChooseOneOrManyStepsOf__a__enabled': True,
            'ChooseOneOrManyStepsOf__b__enabled': False
        },
        hyperparams_space={
            'ChooseOneOrManyStepsOf__a__enabled': Boolean(),
            'ChooseOneOrManyStepsOf__b__enabled': Boolean()
        },
        expected_processed_outputs=np.array([0, 2, 4, 6, 8, 10, 12, 14, 16, 18])
    )


def create_test_case_fit_transform_multiple_steps_choosen():
    a_callback = TapeCallbackFunction()
    b_callback = TapeCallbackFunction()
    c_callback = TapeCallbackFunction()
    d_callback = TapeCallbackFunction()

    return NeuraxleTestCase(
        pipeline=Pipeline([
            ChooseOneOrManyStepsOf([
                ('a', FitTransformCallbackStep(a_callback, c_callback, transform_function=lambda di: di * 2)),
                ('b', FitTransformCallbackStep(b_callback, d_callback, transform_function=lambda di: di * 2))
            ]),
        ]),
        callbacks=[a_callback, c_callback, b_callback, d_callback],
        expected_callbacks_data=[
            DATA_INPUTS,
            (DATA_INPUTS, EXPECTED_OUTPUTS),
            DATA_INPUTS,
            (DATA_INPUTS, EXPECTED_OUTPUTS)
        ],
        hyperparams={
            'ChooseOneOrManyStepsOf__a__enabled': True,
            'ChooseOneOrManyStepsOf__b__enabled': True
        },
        hyperparams_space={
            'ChooseOneOrManyStepsOf__a__enabled': Boolean(),
            'ChooseOneOrManyStepsOf__b__enabled': Boolean()
        },
        expected_processed_outputs=np.array([0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 0, 2, 4, 6, 8, 10, 12, 14, 16, 18])
    )


def create_test_case_fit_single_step_choosen():
    a_callback = TapeCallbackFunction()
    b_callback = TapeCallbackFunction()
    c_callback = TapeCallbackFunction()
    d_callback = TapeCallbackFunction()

    return NeuraxleTestCase(
        pipeline=Pipeline([
            ChooseOneOrManyStepsOf([
                ('a', FitTransformCallbackStep(a_callback, c_callback, transform_function=lambda di: di * 2)),
                ('b', FitTransformCallbackStep(b_callback, d_callback, transform_function=lambda di: di * 2))
            ]),
        ]),
        callbacks=[a_callback, c_callback, b_callback, d_callback],
        expected_callbacks_data=[
            [],
            (DATA_INPUTS, EXPECTED_OUTPUTS),
            [],
            []
        ],
        hyperparams={
            'ChooseOneOrManyStepsOf__a__enabled': True,
            'ChooseOneOrManyStepsOf__b__enabled': False
        },
        hyperparams_space={
            'ChooseOneOrManyStepsOf__a__enabled': Boolean(),
            'ChooseOneOrManyStepsOf__b__enabled': Boolean()
        },
        expected_processed_outputs=np.array([0, 2, 4, 6, 8, 10, 12, 14, 16, 18])
    )


def create_test_case_fit_multiple_steps_choosen():
    a_callback = TapeCallbackFunction()
    b_callback = TapeCallbackFunction()
    c_callback = TapeCallbackFunction()
    d_callback = TapeCallbackFunction()

    return NeuraxleTestCase(
        pipeline=Pipeline([
            ChooseOneOrManyStepsOf([
                ('a', FitTransformCallbackStep(a_callback, c_callback, transform_function=lambda di: di * 2)),
                ('b', FitTransformCallbackStep(b_callback, d_callback, transform_function=lambda di: di * 2))
            ]),
        ]),
        callbacks=[a_callback, c_callback, b_callback, d_callback],
        expected_callbacks_data=[
            [],
            (DATA_INPUTS, EXPECTED_OUTPUTS),
            [],
            (DATA_INPUTS, EXPECTED_OUTPUTS)
        ],
        hyperparams={
            'ChooseOneOrManyStepsOf__a__enabled': True,
            'ChooseOneOrManyStepsOf__b__enabled': True
        },
        hyperparams_space={
            'ChooseOneOrManyStepsOf__a__enabled': Boolean(),
            'ChooseOneOrManyStepsOf__b__enabled': Boolean()
        },
        expected_processed_outputs=np.array([0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 0, 2, 4, 6, 8, 10, 12, 14, 16, 18])
    )


@pytest.mark.parametrize('test_case', [
    create_test_case_fit_transform_single_step_choosen(),
    create_test_case_fit_transform_multiple_steps_choosen()
])
def test_choose_one_or_many_step_of_fit_transform_should_choose_step(
        test_case: NeuraxleTestCase):
    p = test_case.pipeline
    p.set_hyperparams_space(test_case.hyperparams_space)
    p.set_hyperparams(test_case.hyperparams)

    p, outputs = p.fit_transform(DATA_INPUTS, EXPECTED_OUTPUTS)

    assert np.array_equal(outputs, test_case.expected_processed_outputs)
    assert_callback_data_is_as_expected(test_case)


@pytest.mark.parametrize('test_case', [
    create_test_case_fit_single_step_choosen(),
    create_test_case_fit_multiple_steps_choosen()
])
def test_choose_one_or_many_step_of_fit_should_choose_step(
        test_case: NeuraxleTestCase):
    p = test_case.pipeline
    p.set_hyperparams_space(test_case.hyperparams_space)
    p.set_hyperparams(test_case.hyperparams)

    p = p.fit(DATA_INPUTS, EXPECTED_OUTPUTS)

    assert_callback_data_is_as_expected(test_case)


def choose_one_step_single_step_chosen_fit_transform() -> NeuraxleTestCase:
    a_callback = TapeCallbackFunction()
    b_callback = TapeCallbackFunction()
    a_fit_callback = TapeCallbackFunction()
    b_fit_callback = TapeCallbackFunction()

    t = NeuraxleTestCase(
        pipeline=Pipeline([
            ChooseOneStepOf([
                ('a', Pipeline([
                    FitTransformCallbackStep(a_callback, a_fit_callback, transform_function=lambda di: di * 2),
                    AssertFalseStep("passage 'a' should not have been chosen.")
                ])),
                ('b', FitTransformCallbackStep(b_callback, b_fit_callback, transform_function=lambda di: di * 2))
            ]),
        ]),
        callbacks=[a_callback, a_fit_callback, b_callback, b_fit_callback],
        expected_callbacks_data=[
            [],
            [],
            DATA_INPUTS,
            (DATA_INPUTS, EXPECTED_OUTPUTS),
        ],
        hyperparams={
            'ChooseOneStepOf__choice': 'b'
        },
        expected_processed_outputs=np.array([0, 2, 4, 6, 8, 10, 12, 14, 16, 18])
    )
    return t


def choose_one_step_single_step_chosen_fit():
    a_callback = TapeCallbackFunction()
    b_callback = TapeCallbackFunction()
    a_fit_callback = TapeCallbackFunction()
    b_fit_callback = TapeCallbackFunction()

    return NeuraxleTestCase(
        pipeline=Pipeline([
            ChooseOneStepOf([
                ('a', Pipeline([
                    FitTransformCallbackStep(a_callback, a_fit_callback, transform_function=lambda di: di * 2),
                    AssertFalseStep("passage 'a' should not have been chosen.")
                ])),
                ('b', FitTransformCallbackStep(b_callback, b_fit_callback, transform_function=lambda di: di * 2))
            ]),
        ]),
        callbacks=[a_callback, a_fit_callback, b_callback, b_fit_callback],
        expected_callbacks_data=[
            [],
            [],
            [],
            (DATA_INPUTS, EXPECTED_OUTPUTS),
        ],
        hyperparams={
            'ChooseOneStepOf__choice': 'b'
        },
        expected_processed_outputs=np.array([0, 2, 4, 6, 8, 10, 12, 14, 16, 18])
    )


def choose_one_step_single_step_chosen_transform():
    a_callback = TapeCallbackFunction()
    b_callback = TapeCallbackFunction()
    a_fit_callback = TapeCallbackFunction()
    b_fit_callback = TapeCallbackFunction()

    return NeuraxleTestCase(
        pipeline=Pipeline([
            ChooseOneStepOf([
                ('a', Pipeline([
                    FitTransformCallbackStep(a_callback, a_fit_callback, transform_function=lambda di: di * 2),
                    AssertFalseStep("passage 'a' should not have been chosen.")
                ])),
                ('b', FitTransformCallbackStep(b_callback, b_fit_callback, transform_function=lambda di: di * 2))
            ]),
        ]),
        callbacks=[a_callback, a_fit_callback, b_callback, b_fit_callback],
        expected_callbacks_data=[
            [],
            [],
            DATA_INPUTS,
            []
        ],
        hyperparams={
            'ChooseOneStepOf__choice': 'b'
        },
        expected_processed_outputs=np.array([0, 2, 4, 6, 8, 10, 12, 14, 16, 18])
    )


def test_choose_one_step_of_fit_transform_should_choose_step_b():
    test_case: NeuraxleTestCase = choose_one_step_single_step_chosen_fit_transform()
    p = test_case.pipeline
    p.set_hyperparams(test_case.hyperparams)

    p, outputs = p.fit_transform(DATA_INPUTS, EXPECTED_OUTPUTS)

    assert np.array_equal(outputs, test_case.expected_processed_outputs)
    assert_callback_data_is_as_expected(test_case)


def test_choose_one_step_of_fit_should_choose_step_b():
    test_case: NeuraxleTestCase = choose_one_step_single_step_chosen_fit()
    p = test_case.pipeline
    p.set_hyperparams(test_case.hyperparams)

    p = p.fit(DATA_INPUTS, EXPECTED_OUTPUTS)

    assert_callback_data_is_as_expected(test_case)


def test_choose_one_step_of_transform_should_choose_step_b():
    test_case: NeuraxleTestCase = choose_one_step_single_step_chosen_transform()
    p = test_case.pipeline
    p.set_hyperparams(test_case.hyperparams)

    _ = p.transform(DATA_INPUTS)

    assert_callback_data_is_as_expected(test_case)


def test_choose_one_step_of_invalid_chosen_step():
    with pytest.raises(ValueError):
        Pipeline([
            ChooseOneStepOf([
                ('a', Identity()),
                ('b', Identity())
            ]).set_hyperparams({'choice': 'c'}),
        ])


@pytest.mark.parametrize("method_name, args, kwargs", [
    ("set_hyperparams", [{'choice': 'b'}], {}),
    ("update_hyperparams", [{'choice': 'b'}], {}),
    ("apply", ["_update_hyperparams"], {"hyperparams": {'choice': 'b'}})
    # This case correspond to the update in an AutoML loop.
])
def test_choose_one_step_of_set_hyperparams(method_name, args, kwargs):
    a_callback = TapeCallbackFunction()
    b_callback = TapeCallbackFunction()
    c_callback = TapeCallbackFunction()
    d_callback = TapeCallbackFunction()

    choose_one_step_of = ChooseOneStepOf([
        ('a', FitTransformCallbackStep(a_callback, c_callback, transform_function=lambda di: di * 2).set_name("step_1")),
        ('b', FitTransformCallbackStep(b_callback, d_callback, transform_function=lambda di: di * 2).set_name("step_1"))
    ])

    p = Pipeline([
        choose_one_step_of
    ])

    p.transform(DATA_INPUTS)

    assert len(a_callback.data) == 1
    assert all(a_callback.data[0] == DATA_INPUTS)
    assert len(b_callback.data) == 0
    assert len(c_callback.data) == 0
    assert len(d_callback.data) == 0

    getattr(choose_one_step_of, method_name)(*args, **kwargs)

    p.transform(DATA_INPUTS)

    assert len(a_callback.data) == 1
    assert all(a_callback.data[0] == DATA_INPUTS)
    assert len(b_callback.data) == 1
    assert all(b_callback.data[0] == DATA_INPUTS)
    assert len(c_callback.data) == 0
    assert len(d_callback.data) == 0


def test_choose_one_step_of_update_hyperparams():
    a_callback = TapeCallbackFunction()
    b_callback = TapeCallbackFunction()
    c_callback = TapeCallbackFunction()
    d_callback = TapeCallbackFunction()

    choose_one_step_of = ChooseOneStepOf([
        ('a', FitTransformCallbackStep(a_callback, c_callback, transform_function=lambda di: di * 2).set_name("step_1")),
        ('b', FitTransformCallbackStep(b_callback, d_callback, transform_function=lambda di: di * 2).set_name("step_1"))
    ])

    p = Pipeline([
        choose_one_step_of
    ])

    p.transform(DATA_INPUTS)

    assert len(a_callback.data) == 1
    assert all(a_callback.data[0] == DATA_INPUTS)
    assert len(b_callback.data) == 0
    assert len(c_callback.data) == 0
    assert len(d_callback.data) == 0

    choose_one_step_of.update_hyperparams({'choice': 'b'})

    p.transform(DATA_INPUTS)

    assert len(a_callback.data) == 1
    assert all(a_callback.data[0] == DATA_INPUTS)
    assert len(b_callback.data) == 1
    assert all(b_callback.data[0] == DATA_INPUTS)
    assert len(c_callback.data) == 0
    assert len(d_callback.data) == 0



import numpy as np

from neuraxle.base import ExecutionContext as CX
from neuraxle.data_container import DataContainer as DACT
from neuraxle.pipeline import Pipeline
from neuraxle.steps.data import InnerConcatenateDataContainer, ZipBatchDataContainer

TIMESTEPS = 10
FEATURES = 5
VALIDATION_SIZE = 0.1
BATCH_SIZE = 32
N_EPOCHS = 10
SHAPE_3D = (BATCH_SIZE, TIMESTEPS, FEATURES)
SHAPE_2D = (BATCH_SIZE, TIMESTEPS)
SHAPE_1D = BATCH_SIZE


def test_inner_concatenate_data_should_merge_3d_with_3d():
    # Given
    data_inputs_3d, expected_outputs_3d = _create_data_source(SHAPE_3D)
    data_inputs_3d_second, expected_outputs_3d_second = _create_data_source(SHAPE_3D)
    data_container_3d_second = DACT(data_inputs=data_inputs_3d_second,
                                             expected_outputs=expected_outputs_3d_second)
    data_container = DACT(data_inputs=data_inputs_3d, expected_outputs=expected_outputs_3d) \
        .add_sub_data_container('2d', data_container_3d_second)

    # When
    p = Pipeline([
        InnerConcatenateDataContainer(sub_data_container_names=['2d'])
    ])

    data_container = p.handle_transform(data_container, CX())

    # Then
    assert data_container.data_inputs.shape == (SHAPE_3D[0], SHAPE_3D[1], SHAPE_3D[2] * 2)
    assert data_container.expected_outputs.shape == (SHAPE_3D[0], SHAPE_3D[1], SHAPE_3D[2] * 2)
    assert np.array_equal(data_container.data_inputs[..., -SHAPE_3D[2]:], data_container_3d_second.data_inputs)
    assert np.array_equal(data_container.expected_outputs[..., -SHAPE_3D[2]:],
                          data_container_3d_second.expected_outputs)


def test_inner_concatenate_data_should_merge_2d_with_3d():
    # Given
    data_inputs_3d, expected_outputs_3d = _create_data_source(SHAPE_3D)
    data_inputs_2d, expected_outputs_2d = _create_data_source(SHAPE_2D)
    data_container_2d = DACT(data_inputs=data_inputs_2d, expected_outputs=expected_outputs_2d)
    data_container_3d = DACT(data_inputs=data_inputs_3d, expected_outputs=expected_outputs_3d) \
        .add_sub_data_container('2d', data_container_2d)

    # When
    p = Pipeline([
        InnerConcatenateDataContainer(sub_data_container_names=['2d'])
    ])

    data_container_3d = p.handle_transform(data_container_3d, CX())

    # Then
    assert data_container_3d.data_inputs.shape == (SHAPE_3D[0], SHAPE_3D[1], SHAPE_3D[2] + 1)
    assert data_container_3d.expected_outputs.shape == (SHAPE_3D[0], SHAPE_3D[1], SHAPE_3D[2] + 1)
    assert np.array_equal(data_container_3d.data_inputs[..., -1], data_container_2d.data_inputs)
    assert np.array_equal(data_container_3d.expected_outputs[..., -1], data_container_2d.expected_outputs)


def test_inner_concatenate_data_should_merge_1d_with_3d():
    # Given
    data_inputs_3d, expected_outputs_3d = _create_data_source(SHAPE_3D)
    data_inputs_1d, expected_outputs_1d = _create_data_source(SHAPE_1D)
    data_container_1d = DACT(data_inputs=data_inputs_1d, expected_outputs=expected_outputs_1d)
    data_container = DACT(data_inputs=data_inputs_3d, expected_outputs=expected_outputs_3d) \
        .add_sub_data_container('1d', data_container_1d)

    # When
    p = Pipeline([
        InnerConcatenateDataContainer(sub_data_container_names=['1d'])
    ])

    data_container = p.handle_transform(data_container, CX())

    # Then
    broadcasted_data_inputs_1d = np.broadcast_to(np.expand_dims(data_container_1d.data_inputs, axis=-1),
                                                 shape=(SHAPE_3D[0], SHAPE_3D[1]))
    broadcasted_expected_outputs_1d = np.broadcast_to(np.expand_dims(data_container_1d.expected_outputs, axis=-1),
                                                      shape=(SHAPE_3D[0], SHAPE_3D[1]))

    assert np.array_equal(data_container.data_inputs[..., -1], broadcasted_data_inputs_1d)
    assert np.array_equal(data_container.expected_outputs[..., -1], broadcasted_expected_outputs_1d)

    assert data_container.data_inputs.shape == (SHAPE_3D[0], SHAPE_3D[1], SHAPE_3D[2] + 1)
    assert data_container.expected_outputs.shape == (SHAPE_3D[0], SHAPE_3D[1], SHAPE_3D[2] + 1)


def test_inner_concatenate_data_should_merge_1d_with_2d():
    # Given
    data_inputs_2d, expected_outputs_2d = _create_data_source(SHAPE_2D)
    data_inputs_1d, expected_outputs_1d = _create_data_source(SHAPE_1D)
    data_container_1d = DACT(data_inputs=data_inputs_1d, expected_outputs=expected_outputs_1d)
    data_container = DACT(data_inputs=data_inputs_2d, expected_outputs=expected_outputs_2d) \
        .add_sub_data_container('1d', data_container_1d)

    # When
    p = Pipeline([
        InnerConcatenateDataContainer(sub_data_container_names=['1d'])
    ])

    data_container = p.handle_transform(data_container, CX())

    # Then
    assert data_container.data_inputs.shape == (SHAPE_2D[0], SHAPE_2D[1] + 1)
    assert data_container.expected_outputs.shape == (SHAPE_2D[0], SHAPE_2D[1] + 1)
    assert np.array_equal(data_container.data_inputs[..., -1], data_container_1d.data_inputs)
    assert np.array_equal(data_container.expected_outputs[..., -1], data_container_1d.expected_outputs)


def test_outer_concatenate_data_should_merge_2d_with_3d():
    # Given
    data_inputs_3d, expected_outputs_3d = _create_data_source(SHAPE_3D)
    data_inputs_2d, expected_outputs_2d = _create_data_source(SHAPE_2D)
    data_container_2d = DACT(data_inputs=data_inputs_2d, expected_outputs=expected_outputs_2d)
    data_container = DACT(data_inputs=data_inputs_3d, expected_outputs=expected_outputs_3d) \
        .add_sub_data_container('2d', data_container_2d)

    # When
    p = Pipeline([
        ZipBatchDataContainer(sub_data_container_names=['2d'])
    ])

    data_container = p.handle_transform(data_container, CX())

    # Then
    for i, (first_di, second_di) in enumerate(zip(data_inputs_3d, data_inputs_2d)):
        assert np.array_equal(data_container.data_inputs[i][0], first_di)
        assert np.array_equal(data_container.data_inputs[i][1], second_di)

    for i, (first_eo, second_eo) in enumerate(zip(expected_outputs_3d, expected_outputs_2d)):
        assert np.array_equal(data_container.expected_outputs[i][0], first_eo)
        assert np.array_equal(data_container.expected_outputs[i][1], second_eo)


def _create_data_source(shape):
    data_inputs = np.random.random(shape).astype(np.float32)
    expected_outputs = np.random.random(shape).astype(np.float32)
    return data_inputs, expected_outputs



"""
Tests for features steps
========================================

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
import pytest

from neuraxle.hyperparams.space import HyperparameterSamples
from neuraxle.steps.features import Cheap3DTo2DTransformer, FFTPeakBinWithValue
import numpy as np


def test_fft_peak_bin_with_values():
    data_inputs = np.random.random((4, 5, 2))
    step = FFTPeakBinWithValue()

    outputs = step.transform(data_inputs)

    assert outputs.shape == (4, 4)


@pytest.mark.parametrize("hyperparams, expected_feature_count", [
    (HyperparameterSamples({
        'FFT__enabled': True,
        'NumpyMean__enabled': True,
        'NumpyMedian__enabled': True,
        'NumpyMin__enabled': True,
        'NumpyMax__enabled': True
    }), 18),
    (HyperparameterSamples({
        'FFT__enabled': False,
        'NumpyMean__enabled': True,
        'NumpyMedian__enabled': True,
        'NumpyMin__enabled': True,
        'NumpyMax__enabled': True
    }), 8),
    (HyperparameterSamples({
        'FFT__enabled': True,
        'NumpyMean__enabled': False,
        'NumpyMedian__enabled': True,
        'NumpyMin__enabled': True,
        'NumpyMax__enabled': True
    }), 16),
    (HyperparameterSamples({
        'FFT__enabled': True,
        'NumpyMean__enabled': True,
        'NumpyMedian__enabled': False,
        'NumpyMin__enabled': True,
        'NumpyMax__enabled': True
    }), 16),
    (HyperparameterSamples({
        'FFT__enabled': True,
        'NumpyMean__enabled': True,
        'NumpyMedian__enabled': True,
        'NumpyMin__enabled': False,
        'NumpyMax__enabled': True
    }), 16),
    (HyperparameterSamples({
        'FFT__enabled': True,
        'NumpyMean__enabled': True,
        'NumpyMedian__enabled': True,
        'NumpyMin__enabled': True,
        'NumpyMax__enabled': False
    }), 16)
])
def test_cheap_3D_to_2D_transformer(hyperparams: HyperparameterSamples, expected_feature_count: int):
    step = Cheap3DTo2DTransformer()
    step.set_hyperparams(hyperparams=hyperparams)
    data_inputs = np.random.random((7, 5, 2))

    outputs = step.transform(data_inputs)

    assert outputs.shape == (7, expected_feature_count)



from typing import List

import numpy as np
import pytest
from neuraxle.base import ExecutionContext as CX
from neuraxle.data_container import DataContainer as DACT
from neuraxle.hyperparams.distributions import FixedHyperparameter, RandInt
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.metaopt.auto_ml import AutoML, Trainer
from neuraxle.metaopt.callbacks import (BestModelCheckpoint,
                                        EarlyStoppingCallback, MetricCallback,
                                        ScoringCallback)
from neuraxle.metaopt.data.aggregates import MetricResults, Round, Trial
from neuraxle.metaopt.data.json_repo import HyperparamsOnDiskRepository
from neuraxle.metaopt.data.vanilla import AutoMLContext, ScopedLocation
from neuraxle.metaopt.validation import (GridExplorationSampler,
                                         KFoldCrossValidationSplitter,
                                         RandomSearchSampler,
                                         ValidationSplitter)
from neuraxle.pipeline import Pipeline
from neuraxle.steps.flow import ChooseStepElseIdentity
from neuraxle.steps.numpy import MultiplyByN, NumpyReshape
from neuraxle.steps.sklearn import SKLearnWrapper
from sklearn import linear_model
from sklearn.exceptions import ConvergenceWarning
from sklearn.metrics import mean_squared_error
from sklearn.svm import LinearSVC
from sklearn.utils._testing import ignore_warnings


def test_automl_early_stopping_callback(tmpdir):
    # Given
    cx = AutoMLContext.from_context(CX(tmpdir))
    n_epochs = 10
    max_epochs_without_improvement = 3
    auto_ml = AutoML(
        pipeline=Pipeline([
            MultiplyByN(2).set_hyperparams_space(HyperparameterSpace({
                'multiply_by': FixedHyperparameter(2)
            })),
            NumpyReshape(new_shape=(-1, 1)),
        ]),
        hyperparams_optimizer=RandomSearchSampler(),
        validation_splitter=ValidationSplitter(0.20),
        scoring_callback=ScoringCallback(mean_squared_error, higher_score_is_better=False),
        callbacks=[
            MetricCallback('mse', metric_function=mean_squared_error, higher_score_is_better=False),
            EarlyStoppingCallback(max_epochs_without_improvement)
        ],
        n_trials=1,
        refit_best_trial=True,
        epochs=n_epochs,
        continue_loop_on_error=False
    )

    # When
    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    expected_outputs = data_inputs * 2
    auto_ml.handle_fit(DACT(data_inputs=data_inputs, expected_outputs=expected_outputs), cx)

    # Then
    round_scope: Round = cx.with_loc(ScopedLocation.default(0)).load_agg()
    trial: Trial = round_scope.get_best_trial()
    assert len(trial._validation_splits) == 1
    validation_scores = trial._validation_splits[0].metric_result('mse').get_valid_scores()
    nepochs_executed = len(validation_scores)
    assert nepochs_executed == max_epochs_without_improvement + 1


@pytest.mark.skip(reason="https://github.com/Neuraxio/Neuraxle/issues/522")
def test_automl_savebestmodel_callback(tmpdir):
    # Given
    cx = AutoMLContext.from_context(CX(tmpdir))
    validation_splitter = ValidationSplitter(0.20)
    auto_ml = AutoML(
        pipeline=Pipeline([
            MultiplyByN(2).set_hyperparams_space(HyperparameterSpace({
                'multiply_by': FixedHyperparameter(2)
            })),
            NumpyReshape(new_shape=(-1, 1)),
            linear_model.LinearRegression()
        ]),
        validation_splitter=validation_splitter,
        hyperparams_optimizer=RandomSearchSampler(),
        scoring_callback=ScoringCallback(mean_squared_error, higher_score_is_better=False),
        callbacks=[
            BestModelCheckpoint()
        ],
        n_trials=3,
        epochs=1,
        refit_best_trial=False,
        continue_loop_on_error=False
    ).with_context(cx)

    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    expected_outputs = data_inputs * 4

    # When
    auto_ml.fit(data_inputs=data_inputs, expected_outputs=expected_outputs)

    # Then
    trials: Round = cx.with_loc(ScopedLocation.default(0)).load_agg()
    best_trial = trials.get_best_trial()
    best_trial_score = best_trial.get_avg_validation_score()
    best_model = auto_ml.wrapped.wrapped
    _, _, _, valid_inputs, valid_outputs, _ = validation_splitter.split(
        data_inputs=data_inputs, expected_outputs=expected_outputs)
    predicted_output = best_model.predict(*valid_inputs)
    score = mean_squared_error(*valid_outputs, predicted_output)

    assert best_trial_score == score


def test_automl_optional_linreg_mse(tmpdir):
    # Given
    cx = AutoMLContext.from_context(CX(tmpdir))
    auto_ml = AutoML(
        pipeline=Pipeline([
            MultiplyByN(2).with_hp_range(range(1, 5)),
            ChooseStepElseIdentity(
                Pipeline([NumpyReshape(new_shape=(-1, 1)), linear_model.LinearRegression()]),
            )
        ]),
        validation_splitter=ValidationSplitter(0.20),
        hyperparams_optimizer=GridExplorationSampler(10),
        scoring_callback=ScoringCallback(mean_squared_error, higher_score_is_better=False),
        callbacks=[
            MetricCallback('mse', metric_function=mean_squared_error,
                           higher_score_is_better=False),
        ],
        n_trials=10,
        epochs=1,
        refit_best_trial=True,
        continue_loop_on_error=False
    )

    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    expected_outputs = data_inputs * 4

    # When
    auto_ml.handle_fit(DACT(di=data_inputs, eo=expected_outputs), cx)

    # Then
    p = auto_ml.wrapped
    outputs = p.transform(data_inputs)
    mse = mean_squared_error(expected_outputs, outputs)

    assert mse < 1e-10


def test_validation_splitter_should_split_data_properly():
    # Given
    data_inputs = np.random.random((4, 2, 2048, 6)).astype(np.float32)
    expected_outputs = np.random.random((4, 2, 2048, 1)).astype(np.float32)
    splitter = ValidationSplitter(validation_size=0.2)

    # When
    validation_splits = splitter.split_dact(
        data_container=DACT(data_inputs=data_inputs, expected_outputs=expected_outputs),
        context=CX()
    )
    train_di, train_eo, validation_di, validation_eo = extract_validation_split_data(validation_splits)

    train_di = train_di[0]
    train_eo = train_eo[0]

    validation_di = validation_di[0]
    validation_eo = validation_eo[0]

    # Then
    assert len(train_di) == 3
    assert np.array_equal(np.array(train_di), data_inputs[0:3])
    assert len(train_eo) == 3
    assert np.array_equal(np.array(train_eo), expected_outputs[0:3])

    assert len(validation_di) == 1
    assert np.array_equal(validation_di[0], data_inputs[-1])
    assert len(validation_eo) == 1
    assert np.array_equal(validation_eo[0], expected_outputs[-1])


def test_kfold_cross_validation_should_split_data_properly():
    # Given
    data_inputs = np.random.random((4, 2, 2048, 6)).astype(np.float32)
    expected_outputs = np.random.random((4, 2, 2048, 1)).astype(np.float32)
    splitter = KFoldCrossValidationSplitter(k_fold=4)

    # When
    validation_splits = splitter.split_dact(
        data_container=DACT(data_inputs=data_inputs, expected_outputs=expected_outputs),
        context=CX()
    )
    train_di, train_eo, validation_di, validation_eo = extract_validation_split_data(validation_splits)

    # Then
    assert len(train_di[0]) == 3
    assert np.array_equal(np.array(train_di[0]), data_inputs[1:])
    assert len(train_eo[0]) == 3
    assert np.array_equal(np.array(train_eo[0]), expected_outputs[1:])

    assert len(train_di[1]) == 3
    assert np.array_equal(np.array(train_di[1]),
                          np.concatenate((np.expand_dims(data_inputs[0], axis=0), data_inputs[2:]), axis=0))
    assert len(train_eo[1]) == 3
    assert np.array_equal(np.array(train_eo[1]),
                          np.concatenate((np.expand_dims(expected_outputs[0], axis=0), expected_outputs[2:]), axis=0))

    assert len(train_di[2]) == 3
    assert np.array_equal(np.array(train_di[2]),
                          np.concatenate((data_inputs[0:2], np.expand_dims(data_inputs[3], axis=0)), axis=0))
    assert len(train_eo[2]) == 3
    assert np.array_equal(np.array(train_eo[2]),
                          np.concatenate((expected_outputs[0:2], np.expand_dims(expected_outputs[3], axis=0)), axis=0))

    assert len(train_di[3]) == 3
    assert np.array_equal(np.array(train_di[3]), data_inputs[0:3])
    assert len(train_eo[3]) == 3
    assert np.array_equal(np.array(train_eo[3]), expected_outputs[0:3])

    assert len(validation_di[0]) == 1
    assert np.array_equal(validation_di[0][0], data_inputs[0])
    assert len(validation_eo[0]) == 1
    assert np.array_equal(validation_eo[0][0], expected_outputs[0])

    assert len(validation_di[1]) == 1
    assert np.array_equal(validation_di[1][0], data_inputs[1])
    assert len(validation_eo[1]) == 1
    assert np.array_equal(validation_eo[1][0], expected_outputs[1])

    assert len(validation_di[2]) == 1
    assert np.array_equal(validation_di[2][0], data_inputs[2])
    assert len(validation_eo[2]) == 1
    assert np.array_equal(validation_eo[2][0], expected_outputs[2])

    assert len(validation_di[3]) == 1
    assert np.array_equal(validation_di[3][0], data_inputs[3])
    assert len(validation_eo[3]) == 1
    assert np.array_equal(validation_eo[3][0], expected_outputs[3])


def test_kfold_cross_validation_should_split_data_properly_bug():
    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    expected_outputs = np.array([0, 4, 8, 12, 16, 20, 24, 28, 32, 36, 40])
    data_container = DACT(
        data_inputs=data_inputs,
        expected_outputs=expected_outputs
    )
    splitter = KFoldCrossValidationSplitter(k_fold=2)

    # When
    validation_splits = splitter.split_dact(data_container, CX())

    train_di, train_eo, validation_di, validation_eo = extract_validation_split_data(validation_splits)

    # Then
    assert len(train_di[0]) == 6
    assert np.array_equal(np.array(train_di[0]), data_inputs[5:])
    assert len(train_eo[0]) == 6
    assert np.array_equal(np.array(train_eo[0]), expected_outputs[5:])

    assert len(train_di[1]) == 5
    assert np.array_equal(
        np.array(train_di[1]),
        data_inputs[:5]
    )
    assert len(train_eo[1]) == 5
    assert np.array_equal(
        np.array(train_eo[1]),
        expected_outputs[:5]
    )

    assert len(validation_di[0]) == 5
    assert np.array_equal(np.array(validation_di[0]), data_inputs[:5])
    assert len(validation_eo[0]) == 5
    assert np.array_equal(np.array(validation_eo[0]), expected_outputs[:5])

    assert len(validation_di[1]) == 6
    assert np.array_equal(np.array(validation_di[1]), data_inputs[5:])
    assert len(validation_eo[1]) == 6
    assert np.array_equal(validation_eo[1], expected_outputs[5:])


def extract_validation_split_data(validation_splits):
    train_di = []
    train_eo = []
    validation_di = []
    validation_eo = []
    for train_dc, validation_dc in validation_splits:
        train_di.append(train_dc.data_inputs)
        train_eo.append(train_dc.expected_outputs)

        validation_di.append(validation_dc.data_inputs)
        validation_eo.append(validation_dc.expected_outputs)
    return train_di, train_eo, validation_di, validation_eo


@ignore_warnings(category=ConvergenceWarning)
def test_automl_should_shallow_copy_data_before_each_epoch(tmpdir):
    # see issue #332 https://github.com/Neuraxio/Neuraxle/issues/332
    data_inputs = np.random.randint(0, 100, (100, 3))
    expected_outputs = np.array(np.sum(data_inputs, axis=-1) / 100, dtype=int)
    data_inputs += np.random.randint(0, 100, (100, 3))

    from sklearn.preprocessing import StandardScaler
    p = Pipeline([
        SKLearnWrapper(StandardScaler()),
        SKLearnWrapper(LinearSVC(), HyperparameterSpace({'C': RandInt(0, 10000)})),
    ])

    auto_ml = AutoML(
        p,
        validation_splitter=ValidationSplitter(0.20),
        refit_best_trial=True,
        n_trials=10,
        epochs=1,
        scoring_callback=ScoringCallback(
            name='mse',
            metric_function=mean_squared_error,
            higher_score_is_better=False),
        continue_loop_on_error=False
    )

    random_search = auto_ml.fit(data_inputs, expected_outputs)

    best_model = random_search.wrapped

    assert isinstance(best_model, Pipeline)


def test_trainer_train():
    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    expected_outputs = data_inputs * 4
    dact = DACT(di=data_inputs, eo=expected_outputs)
    ts: Trial = Trial.dummy()
    p = Pipeline([
        MultiplyByN(2).set_hyperparams_space(HyperparameterSpace({
            'multiply_by': FixedHyperparameter(2)
        })),
        NumpyReshape(new_shape=(-1, 1)),
        linear_model.LinearRegression()
    ])

    trainer: Trainer = Trainer(
        n_epochs=10,
        callbacks=[ScoringCallback(mean_squared_error, higher_score_is_better=False)],
        validation_splitter=ValidationSplitter(validation_size=0.20)
    )

    trainer.train(p, dact, ts)
    trained_pipeline = trainer.refit(p, dact, ts)

    outputs = trained_pipeline.transform(data_inputs * 3)
    mse = mean_squared_error(expected_outputs * 3, outputs)

    assert mse < 0.01



import numpy as np

from neuraxle.base import ExecutionContext as CX
from neuraxle.data_container import DataContainer as DACT
from neuraxle.pipeline import Pipeline
from neuraxle.steps.flow import TestOnlyWrapper, TrainOnlyWrapper
from neuraxle.steps.misc import TapeCallbackFunction, CallbackWrapper
from neuraxle.steps.numpy import MultiplyByN

from testing.mocks.step_mocks import SomeStepWithHyperparams


def test_basestep_print_str_representation_works_correctly():
    output = str(SomeStepWithHyperparams())
    assert output == "SomeStepWithHyperparams(name='MockStep')"


def test_basestep_repr_representation_works_correctly():
    output = repr(SomeStepWithHyperparams())
    assert output == """SomeStepWithHyperparams(name='MockStep', hyperparameters=HyperparameterSamples([('learning_rate', 0.1),
                       ('l2_weight_reg', 0.001),
                       ('hidden_size', 32),
                       ('num_layers', 3),
                       ('num_lstm_layers', 1),
                       ('use_xavier_init', True),
                       ('use_max_pool_else_avg_pool', True),
                       ('dropout_drop_proba', 0.5),
                       ('momentum', 0.1)]))"""


def test_handle_predict_should_predict_in_test_mode():
    tape_fit = TapeCallbackFunction()
    tape_transform = TapeCallbackFunction()
    p = Pipeline([
        TestOnlyWrapper(CallbackWrapper(MultiplyByN(2), tape_transform, tape_fit)),
        TrainOnlyWrapper(CallbackWrapper(MultiplyByN(4), tape_transform, tape_fit))
    ])

    data_container = p.handle_predict(
        data_container=DACT(data_inputs=np.array([1, 1]), expected_outputs=np.array([1, 1])),
        context=CX()
    )

    assert np.array_equal(data_container.data_inputs, np.array([2, 2]))


def test_handle_predict_should_handle_transform_with_initial_is_train_mode_after_predict():
    tape_fit = TapeCallbackFunction()
    tape_transform = TapeCallbackFunction()
    p = Pipeline([
        TestOnlyWrapper(CallbackWrapper(MultiplyByN(2), tape_transform, tape_fit)),
        TrainOnlyWrapper(CallbackWrapper(MultiplyByN(4), tape_transform, tape_fit))
    ])
    data_container = DACT(data_inputs=np.array([1, 1]), expected_outputs=np.array([1, 1]))

    p.handle_predict(
        data_container=data_container.copy(),
        context=CX()
    )
    data_container = p.handle_transform(data_container, CX())

    assert np.array_equal(data_container.data_inputs, np.array([4, 4]))


def test_predict_should_predict_in_test_mode():
    tape_fit = TapeCallbackFunction()
    tape_transform = TapeCallbackFunction()
    p = Pipeline([
        TestOnlyWrapper(CallbackWrapper(MultiplyByN(2), tape_transform, tape_fit)),
        TrainOnlyWrapper(CallbackWrapper(MultiplyByN(4), tape_transform, tape_fit))
    ])

    outputs = p.predict(np.array([1, 1]))

    assert np.array_equal(outputs, np.array([2, 2]))


def test_predict_should_transform_with_initial_is_train_mode_after_predict():
    tape_fit = TapeCallbackFunction()
    tape_transform = TapeCallbackFunction()
    p = Pipeline([
        TestOnlyWrapper(CallbackWrapper(MultiplyByN(2), tape_transform, tape_fit)),
        TrainOnlyWrapper(CallbackWrapper(MultiplyByN(4), tape_transform, tape_fit))
    ])

    p.predict(np.array([1, 1]))
    outputs = p.transform(np.array([1, 1]))

    assert np.array_equal(outputs, np.array([4, 4]))



"""
Neuraxle's Pipeline Classes
====================================
This is the core of Neuraxle's pipelines. You can chain steps to call them one after an other.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
import warnings
from abc import ABC, abstractmethod
from typing import Any, List, Tuple, Union

from neuraxle.base import BaseStep
from neuraxle.base import ExecutionContext as CX
from neuraxle.base import (ExecutionMode, ForceHandleMixin, Identity,
                           NamedStepsList, TruncableSteps,
                           _CustomHandlerMethods)
from neuraxle.data_container import AbsentValuesNullObject
from neuraxle.data_container import DataContainer as DACT
from neuraxle.data_container import ListDataContainer, ZipDataContainer
from neuraxle.logging.warnings import warn_deprecated_arg


class BasePipeline(TruncableSteps, ABC):
    """
    Pipeline is a list of steps. This base class is the base class for all pipelines.

    .. seealso::
        :class:`~neuraxle.pipeline.Pipeline`,
        :class:`~neuraxle.pipeline.MiniBatchSequentialPipeline`,
        :class:`~neuraxle.distributed.streaming.SequentialQueuedPipeline`
    """

    def __init__(self, steps: NamedStepsList):
        TruncableSteps.__init__(self, steps_as_tuple=steps)

    @abstractmethod
    def fit(self, data_inputs, expected_outputs=None) -> 'BasePipeline':
        raise NotImplementedError()

    @abstractmethod
    def transform(self, data_inputs):
        raise NotImplementedError()

    @abstractmethod
    def fit_transform(self, data_inputs, expected_outputs=None) -> Tuple['BasePipeline', Any]:
        raise NotImplementedError()


class Pipeline(BasePipeline):
    """
    Handle methods are used to handle the fit, transform, fit_transform and inverse_transform methods
    using a pipe and filter design pattern to handle the steps sequentially. Each step is called
    with the previous step's output as input. The last step's output is returned as final output.

    Think like the termina pipe commands chained together with the character '|', but where each
    command is a step, and where each step can acquire some state from fitting the transformed
    output of the previous step. This way, the steps can be chained together and the state of each
    step can be saved and reused in the end.

    When a pipeline fit, a new pipeline is returned containing the fitted steps within itself.

    Also check out the parallelized version of the pipeline,
    :class:`~neuraxle.distributed.streaming.SequentialQueuedPipeline`.

    .. seealso::
        :class:`~neuraxle.pipeline.BasePipeline`,
        :class:`~neuraxle.pipeline.MiniBatchSequentialPipeline`
        :class:`~neuraxle.distributed.streaming.SequentialQueuedPipeline`
    """

    def __init__(self, steps: NamedStepsList):
        BasePipeline.__init__(self, steps=steps)

    def fit(self, data_inputs, expected_outputs=None) -> 'Pipeline':
        """
        After loading the last checkpoint, fit each pipeline steps

        :param data_inputs: the data input to fit on
        :param expected_outputs: the expected data output to fit on
        :return: the pipeline itself
        """
        return self.fit_data_container(DACT(di=data_inputs, eo=expected_outputs))

    def transform(self, data_inputs: Any):
        """
        Transform each pipeline steps with the pipe and filter design pattern.

        :param data_inputs: the data input to transform
        :return: transformed data inputs
        """
        data_container = self.transform_data_container(DACT(di=data_inputs))
        return data_container.data_inputs

    def fit_transform(self, data_inputs, expected_outputs=None) -> Tuple['Pipeline', Any]:
        """
        After loading the last checkpoint, fit transform each pipeline steps

        :param data_inputs: the data input to fit on
        :param expected_outputs: the expected data output to fit on
        :return: the pipeline itself
        """
        new_self, data_container = self.fit_transform_data_container(
            DACT(di=data_inputs, eo=expected_outputs))
        return new_self, data_container.data_inputs

    def inverse_transform(self, processed_outputs) -> Any:
        """
        After transforming all data inputs, and obtaining a prediction, we can inverse transform the processed outputs

        :param processed_outputs: the forward transformed data input
        :return: backward transformed processed outputs
        """
        warn_deprecated_arg(self, "handler method", "handle_inverse_transform",
                            "inverse_transform", "handle_inverse_transform", Tuple[DACT, CX])
        context = CX(execution_mode=ExecutionMode.INVERSE_TRANSFORM)
        data_container = DACT(data_inputs=processed_outputs)
        for step in list(reversed(self.values())):
            data_container = step.handle_inverse_transform(data_container, context)
        return data_container.data_inputs

    def fit_data_container(self, data_container) -> 'Pipeline':
        context = CX(execution_mode=ExecutionMode.FIT)
        return self.handle_fit(data_container, context)

    def transform_data_container(self, data_container: DACT):
        context = CX(execution_mode=ExecutionMode.TRANSFORM)
        data_container = self.handle_transform(data_container, context)
        return data_container

    def fit_transform_data_container(self, data_container) -> Tuple['Pipeline', DACT]:
        context = CX(execution_mode=ExecutionMode.FIT_TRANSFORM)
        new_self, data_container = self.handle_fit_transform(data_container, context)
        return new_self, data_container

    def _fit_data_container(self, data_container: DACT, context: CX) -> 'Pipeline':
        """
        After loading the last checkpoint, fit transform each pipeline steps,
        but only fit the last pipeline step.

        :param data_container: the data container to fit transform on
        :param context: execution context
        :return: tuple(pipeline, data_container)
        """
        index_last_step = len(self.steps) - 1
        new_steps_as_tuple: NamedStepsList = []

        for index, (step_name, step) in enumerate(self.steps_as_tuple):
            if index != index_last_step:
                step, data_container = step.handle_fit_transform(data_container, context)
            else:
                step = step.handle_fit(data_container, context)

            new_steps_as_tuple.append((step_name, step))
        self.set_steps(new_steps_as_tuple)
        return self

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        """
        After loading the last checkpoint, transform each pipeline steps

        :param data_container: the data container to transform
        :return: transformed data container
        """
        for _, step in self.steps_as_tuple:
            data_container = step.handle_transform(data_container, context)
        return data_container

    def _fit_transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple['Pipeline', DACT]:
        """
        After loading the last checkpoint, fit transform each pipeline steps

        :param data_container: the data container to fit transform on
        :param context: execution context
        :return: tuple(pipeline, data_container)
        """
        new_steps_as_tuple: NamedStepsList = []

        for step_name, step in self.steps_as_tuple:
            step, data_container = step.handle_fit_transform(data_container, context)
            new_steps_as_tuple.append((step_name, step))

        self.set_steps(new_steps_as_tuple)
        return self, data_container

    def _inverse_transform_data_container(
            self, data_container: DACT, context: CX) -> DACT:
        """
        After transforming all data inputs, and obtaining a prediction, we can inverse transform the processed outputs
        """
        for step_name, step in list(reversed(self.items())):
            data_container = step.handle_inverse_transform(data_container, context)
        return data_container


class MiniBatchSequentialPipeline(_CustomHandlerMethods, ForceHandleMixin, Pipeline):
    """
    Mini Batch Sequential Pipeline class to create a pipeline processing data inputs in batch.

    Provide a default batch size :

    .. code-block:: python

        data_inputs = np.array(list(range(10)))
        pipeline = MiniBatchSequentialPipeline([
            SomeStep()
        ], batch_size=2)
        pipeline.transform(data_inputs)
        # SomeStep will receive [array([0, 1]), array([2, 3]), ..., array([8, 9])]

        pipeline = MiniBatchSequentialPipeline([
            SomeStep()
        ], batch_size=3, keep_incomplete_batch=False)
        pipeline.transform(data_inputs)
        # SomeStep will receive: [array([0, 1, 2]), array([3, 4, 5]), array([6, 7, 8])]

        pipeline = MiniBatchSequentialPipeline(
            [SomeStep()],
            batch_size=3,
            keep_incomplete_batch=True,
            keep_incomplete_batch=True,
            default_value_data_inputs=None,
            default_value_expected_outputs=None
        )
        pipeline.transform(data_inputs)
        # SomeStep will receive: [array([0, 1, 2]), array([3, 4, 5]), array([6, 7, 8]), array([9, None, None])]

        pipeline = MiniBatchSequentialPipeline(
            [SomeStep()],
            batch_size=3,
            keep_incomplete_batch=True,
            default_value_data_inputs=AbsentValuesNullObject()
        )
        pipeline.transform(data_inputs)
        # SomeStep will receive: [array([0, 1, 2]), array([3, 4, 5]), array([6, 7, 8]), array([9])]


    Or manually add one or multiple :class`Barrier` steps to the mini batch sequential pipeline :

    .. code-block:: python

        data_inputs = np.array(list(range(10)))
        pipeline = MiniBatchSequentialPipeline([
            SomeStep(),
            Joiner(batch_size=2)
        ])
        pipeline.transform(data_inputs)
        # SomeStep will receive [array([0, 1]), array([2, 3]), ..., array([8, 9])]

        pipeline = MiniBatchSequentialPipeline([
            SomeStep(),
            Joiner(batch_size=3, keep_incomplete_batch=False)
        ])
        pipeline.transform(data_inputs)
        # SomeStep will receive: [array([0, 1, 2]), array([3, 4, 5]), array([6, 7, 8])]

        pipeline = MiniBatchSequentialPipeline([
            SomeStep(),
            Joiner(
                batch_size=3,
                keep_incomplete_batch=True,
                keep_incomplete_batch=True,
                default_value_data_inputs=None,
                default_value_expected_outputs=None
            )
        ])
        pipeline.transform(data_inputs)
        # SomeStep will receive: [array([0, 1, 2]), array([3, 4, 5]), array([6, 7, 8]), array([9, None, None])]

        pipeline = MiniBatchSequentialPipeline([
            SomeStep(),
            Joiner(
                batch_size=3,
                keep_incomplete_batch=True,
                default_value_data_inputs=AbsentValuesNullObject()
            )
        ])
        pipeline.transform(data_inputs)
        # SomeStep will receive: [array([0, 1, 2]), array([3, 4, 5]), array([6, 7, 8]), array([9])]


    :param steps: pipeline steps
    :param batch_size: number of elements to combine into a single batch
    :param keep_incomplete_batch: (Optional.) A bool representing
    whether the last batch should be dropped in the case it has fewer than
    `batch_size` elements; the default behavior is not to drop the smaller
    batch.
    :param default_value_data_inputs: expected_outputs default fill value
    for padding and values outside iteration range, or :class:`~neuraxle.data_container.DataContainer.AbsentValuesNullObject`
    to trim absent values from the batch
    :param default_value_expected_outputs: expected_outputs default fill value
    for padding and values outside iteration range, or :class:`~neuraxle.data_container.DataContainer.AbsentValuesNullObject`
    to trim absent values from the batch
    :param cache_folder: cache_folder if its at the root of the pipeline
    :param mute_joiner_batch_size_warning: If False, will log a warning when automatically setting the joiner batch_size attribute.

    .. seealso::
        :func:`~neuraxle.data_container.DataContainer.minibatches`,
        :class:`~neuraxle.data_container.DataContainer.AbsentValuesNullObject`,
        :class:`Pipeline`,
        :class:`Barrier`,
        :class:`Joiner`,
        :class:`~neuraxle.data_container.DataContainer`,
        :class:`~neuraxle.base.ExecutionContext`,
        :class:`~neuraxle.distributed.streaming.SequentialQueuedPipeline`

    """

    def __init__(
            self,
            steps: NamedStepsList,
            batch_size=None,
            keep_incomplete_batch: bool = None,
            default_value_data_inputs=AbsentValuesNullObject(),
            default_value_expected_outputs=None,
            mute_joiner_batch_size_warning: bool = True
    ):
        Pipeline.__init__(self, steps=steps)
        ForceHandleMixin.__init__(self)
        self.default_value_data_inputs = default_value_data_inputs
        self.default_value_expected_outputs = default_value_expected_outputs
        self.__validate_barriers_batch_size(batch_size=batch_size)
        self.__patch_missing_barrier(
            batch_size=batch_size,
            keep_incomplete_batch=keep_incomplete_batch,
            default_value_data_inputs=default_value_data_inputs,
            default_value_expected_outputs=default_value_expected_outputs
        )
        self.mute_joiner_batch_size_warning = mute_joiner_batch_size_warning
        self.__patch_barriers_batch_size(batch_size)

    def set_batch_size(self, batch_size):
        self.__patch_barriers_batch_size(batch_size)

    def __validate_barriers_batch_size(self, batch_size):
        if batch_size is not None:
            return

        for _, step in self:
            if isinstance(step, Barrier):
                if step.batch_size is None:
                    raise Exception(
                        'Invalid Joiner batch size {}[{}]. Please provide a default batch size to MiniBatchSequentialPipeline, or add a batch size to {}[{}].'.format(
                            self.name, step.name, self.name, step.name))

    def __patch_barriers_batch_size(self, batch_size):
        if batch_size is None:
            return

        for _, step in self:
            if isinstance(step, Barrier):
                if step.batch_size is not None and not self.mute_joiner_batch_size_warning:
                    warnings.warn(
                        'Replacing {}[{}].batch_size by {}.batch_size.'.format(self.name, step.name, self.name))
                step.batch_size = batch_size

    def __patch_missing_barrier(
            self,
            batch_size: int,
            keep_incomplete_batch: bool,
            default_value_data_inputs: Union[Any, AbsentValuesNullObject] = None,
            default_value_expected_outputs: Union[Any, AbsentValuesNullObject] = None
    ):
        has_barrier: bool = False

        for _, step in self:
            if isinstance(step, Barrier):
                has_barrier = True

        if not has_barrier:
            self.steps_as_tuple.append((
                'Joiner',
                Joiner(
                    batch_size=batch_size,
                    keep_incomplete_batch=keep_incomplete_batch,
                    default_value_data_inputs=default_value_data_inputs,
                    default_value_expected_outputs=default_value_expected_outputs
                )
            ))

        self._refresh_steps()

    def transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        """
        Transform all sub pipelines splitted by the Barrier steps.
        :param data_container: data container to transform.
        :param context: execution context
        :return: data container
        """
        sub_pipelines = self._create_sub_pipelines()

        for sub_pipeline in sub_pipelines:
            barrier = sub_pipeline[-1]
            data_container = barrier.join_transform(
                step=sub_pipeline,
                data_container=data_container,
                context=context
            )

        return data_container

    def fit_data_container(self, data_container: DACT, context: CX) -> BaseStep:
        """
        Fit all sub pipelines splitted by the Barrier steps.
        :param data_container: data container to transform.
        :param context: execution context
        :return: data container
        """
        self, data_container = self.fit_transform_data_container(data_container, context)
        return self

    def fit_transform_data_container(self, data_container: DACT, context: CX) -> Tuple[
            BaseStep, DACT]:
        """
        Transform all sub pipelines splitted by the Barrier steps.
        :param data_container: data container to transform.
        :param context: execution context
        :return: data container
        """
        sub_pipelines = self._create_sub_pipelines()
        index_start = 0

        for sub_pipeline in sub_pipelines:
            sub_pipeline._setup(context=context)

            barrier = sub_pipeline[-1]
            sub_pipeline, data_container = barrier.join_fit_transform(
                step=sub_pipeline,
                data_container=data_container,
                context=context
            )

            new_self = self[:index_start] + sub_pipeline
            if index_start + len(sub_pipeline) < len(self):
                new_self += self[index_start + len(sub_pipeline):]

            self.steps_as_tuple = new_self.steps_as_tuple
            index_start += len(sub_pipeline)

        return self, data_container

    def _create_sub_pipelines(self) -> List['MiniBatchSequentialPipeline']:
        """
        Create sub pipelines by splitting the steps by the join type name.
        :return: list of sub pipelines
        """
        sub_pipelines: List[MiniBatchSequentialPipeline] = self.split(Barrier)
        for sub_pipeline in sub_pipelines:
            if not sub_pipeline.ends_with(Barrier):
                raise Exception('At least one Barrier step needs to be at the end of a streaming pipeline.')

        return sub_pipelines


class Barrier(Identity, ABC):
    """
    A Barrier step to be used in a minibatch sequential pipeline. It forces all the
    data inputs to get to the barrier in a sub pipeline before going through to the next sub-pipeline.

    .. code-block:: python

        p = MiniBatchSequentialPipeline([
            SomeStep(),
            SomeStep(),
            Barrier(), # must be a concrete Barrier ex: Joiner()
            SomeStep(),
            SomeStep(),
            Barrier(), # must be a concrete Barrier ex: Joiner()
        ], batch_size=10)


    .. seealso::
        :class:`~neuraxle.base.NonTransformableMixin`,
        :class:`~neuraxle.base.BaseStep`
    """

    @abstractmethod
    def join_transform(self, step: TruncableSteps, data_container: DACT,
                       context: CX) -> DACT:
        """
        Execute the given pipeline :func:`~neuraxle.pipeline.Pipeline.transform` with the given data container, and execution context.
        :param step: truncable steps to execute
        :type step: TruncableSteps
        :param data_container: data container
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: transformed data container
        :rtype: DataContainer
        """
        raise NotImplementedError()

    @abstractmethod
    def join_fit_transform(self, step: Pipeline, data_container: DACT, context: CX) -> Tuple[
            'Any', DACT]:
        """
        Execute the given pipeline :func:`~neuraxle.pipeline.Pipeline.fit_transform` with the given data container, and execution context.
        :param step: truncable steps to execute
        :param data_container: data container
        :param context: execution context
        :return: (fitted step, transformed data container)
        """
        raise NotImplementedError()


class Joiner(Barrier):
    """
    A Special Barrier step that joins the transformed mini batches together with list.extend method.

    .. seealso::
        :class:`~neuraxle.data_container.DataContainer`,
        :func:`~neuraxle.data_container.DataContainer.batch`
    """

    def __init__(
            self,
            batch_size: int,
            keep_incomplete_batch: bool = True,
            default_value_data_inputs=AbsentValuesNullObject(),
            default_value_expected_outputs=None
    ):
        Barrier.__init__(self)
        self.batch_size: int = batch_size
        self.keep_incomplete_batch: bool = keep_incomplete_batch
        self.default_value_data_inputs: Union[Any, AbsentValuesNullObject] = default_value_data_inputs
        self.default_value_expected_outputs: Union[Any, AbsentValuesNullObject] = default_value_expected_outputs

    def join_transform(self, step: Pipeline, data_container: DACT, context: CX) -> DACT:
        """
        Concatenate the pipeline transform output of each batch of self.batch_size together.
        :param step: pipeline to transform on
        :type step: Pipeline
        :param data_container: data container to transform
        :type data_container: DataContainer
        :param context: execution context
        :return: transformed data container
        :rtype: DataContainer
        """
        context = context.push(step)
        data_container_batches = data_container.minibatches(
            batch_size=self.batch_size,
            keep_incomplete_batch=self.keep_incomplete_batch,
            default_value_data_inputs=self.default_value_data_inputs,
            default_value_expected_outputs=self.default_value_expected_outputs
        )

        output_data_container = ListDataContainer.empty()
        for data_container_batch in data_container_batches:
            output_data_container.concat(step._transform_data_container(data_container_batch, context))

        return output_data_container

    def join_fit_transform(self, step: Pipeline, data_container: DACT, context: CX) -> \
            Tuple['Any', DACT]:
        """
        Concatenate the pipeline fit transform output of each batch of self.batch_size together.
        :param step: pipeline to fit transform on
        :type step: Pipeline
        :param data_container: data container to fit transform on
        :type data_container: DataContainer
        :param context: execution context
        :return: fitted self, transformed data inputs
        :rtype: Tuple[Any, DataContainer]
        """
        context = context.push(step)
        data_container_batches = data_container.minibatches(
            batch_size=self.batch_size,
            keep_incomplete_batch=self.keep_incomplete_batch,
            default_value_data_inputs=self.default_value_data_inputs,
            default_value_expected_outputs=self.default_value_expected_outputs
        )

        output_data_container = ListDataContainer.empty()
        for data_container_batch in data_container_batches:
            step, data_container_batch = step._fit_transform_data_container(data_container_batch, context)
            output_data_container.concat(data_container_batch)

        return step, output_data_container


class ZipMinibatchJoiner(Joiner):
    """
    Zips together minibatch outputs, i.e. returns a DataContainer where the first
    element is a tuple of every minibatches first element and so on.
    """

    def join_transform(self, step: TruncableSteps, data_container: DACT,
                       context: CX) -> ZipDataContainer:
        context = context.push(step)
        data_container_batches = data_container.minibatches(
            batch_size=self.batch_size,
            keep_incomplete_batch=self.keep_incomplete_batch,
            default_value_data_inputs=self.default_value_data_inputs,
            default_value_expected_outputs=self.default_value_expected_outputs
        )

        output_data_container = []
        for data_container_batch in data_container_batches:
            output_data_container.append(step._transform_data_container(data_container_batch, context))

        return ZipDataContainer.create_from(*output_data_container)

    def join_fit_transform(self, step: Pipeline, data_container: DACT, context: CX) -> \
            Tuple['Any', DACT]:
        context = context.push(step)
        data_container_batches = data_container.minibatches(
            batch_size=self.batch_size,
            keep_incomplete_batch=self.keep_incomplete_batch,
            default_value_data_inputs=self.default_value_data_inputs,
            default_value_expected_outputs=self.default_value_expected_outputs
        )

        output_data_container = []
        for data_container_batch in data_container_batches:
            step, data_container_batch = step._fit_transform_data_container(data_container_batch, context)
            output_data_container.append(data_container_batch)

        return step, ZipDataContainer.create_from(*output_data_container)



import math
from abc import ABC, abstractmethod
from typing import List

import numpy as np
from neuraxle.hyperparams.distributions import (
    ContinuousHyperparameterDistribution, DiscreteHyperparameterDistribution,
    LogSpaceDistributionMixin)
from scipy.special import factorial
from scipy.stats import rv_continuous, rv_discrete, rv_histogram


def scipy_method(func):
    def wrapper(*args, **kwargs):
        self_in_args: ScipyDistributionWrapper = args[0]
        self_in_args._override_scipy_methods()
        return func(*args, **kwargs)

    return wrapper


class ScipyDistributionWrapper(ABC):
    """
    Base class for a distribution that wraps a scipy distribution.

    Usage example:

    .. code-block:: python

        distribution = ScipyDistributionWrapper(
            scipy_distribution=rv_histogram(histogram=histogram),
            null_default_value=null_default_value
        )

    .. seealso::
        :class:`HyperparameterDistribution`
    """

    def __init__(self, scipy_distribution, **scipy_distribution_arguments):
        self.scipy_distribution = scipy_distribution
        self.scipy_distribution_arguments = scipy_distribution_arguments

    def _override_scipy_methods(self):
        return

    @scipy_method
    def rvs(self, *args, **kwargs) -> float:
        return self.scipy_distribution.rvs(*args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def rvs_many(self, size: int, *args, **kwargs) -> List:
        return self.scipy_distribution.rvs(size=size, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def pdf(self, x, *args, **kwargs) -> float:
        if hasattr(self.scipy_distribution, 'pdf'):
            return self.scipy_distribution.pdf(x, *args, **kwargs, **self.scipy_distribution_arguments)
        else:
            return self.scipy_distribution.pmf(x, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def cdf(self, x, *args, **kwargs) -> float:
        """
        Cumulative distribution function of the given x.

        :param x:
        :param args:
        :param kwargs:
        :return:
        """
        return self.scipy_distribution.cdf(x, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def entropy(self, *args, **kwargs):
        """
        Differential entropy of the RV.

        :return:
        """
        return self.scipy_distribution.entropy(*args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def expect(self, *args, **kwargs):
        """
        Calculate expected value of a function with respect to the distribution by numerical integration.

        :param args:
        :param kwargs:
        :return:
        """
        return self.scipy_distribution.expect(*args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def fit(self, data, *args, **kwargs):
        """
        Return MLEs for shape( if applicable), location, and scale parameters from data.

        :param data:
        :param args:
        :param kwargs:
        :return:
        """
        return self.scipy_distribution.fit(data, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def fit_loc_scale(self, data, *args):
        """
        Estimate loc and scale parameters from data using 1st and 2nd moments.
        """
        return self.scipy_distribution.fit_loc_scale(data, *args, **self.scipy_distribution_arguments)

    @scipy_method
    def freeze(self, *args, **kwargs):
        """
        Freeze the distribution for the given arguments.

        :return:
        """
        return self.scipy_distribution.freeze(*args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def interval(self, alpha, *args, **kwargs):
        """
        Confidence interval with equal areas around the median.

        :param alpha:
        :return:
        """
        return self.scipy_distribution.interval(alpha, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def isf(self, q, *args, **kwargs):
        """
        Inverse survival function(inverse of sf) at q of the given RV.

        :param q:
        :return:
        """
        return self.scipy_distribution.isf(q, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def logcdf(self, x, *args, **kwargs):
        """
        Log of the cumulative distribution function at x of the given RV.

        :param x:
        :param args:
        :param kwargs:
        :return:
        """
        return self.scipy_distribution.logcdf(x, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def logpdf(self, x, *args, **kwargs):
        """
        Log of the probability density function at x of the given RV.

        :param x:
        :param args:
        :param kwargs:
        :return:
        """
        return self.scipy_distribution.logpdf(x, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def logsf(self, x, *args, **kwargs):
        """
        Log of the survival function of the given RV.

        :param x:
        :return:
        """
        return self.scipy_distribution.logsf(x, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def min(self):
        return self.scipy_distribution.a

    @scipy_method
    def max(self):
        return self.scipy_distribution.b

    @scipy_method
    def mean(self, *args, **kwargs):
        """
        Mean of the distribution.

        :return:
        """
        return self.scipy_distribution.mean(*args, **kwargs)

    @scipy_method
    def median(self, *args, **kwargs):
        """
        Median of the distribution.

        :return:
        """
        return self.scipy_distribution.median(*args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def moment(self, n, *args, **kwargs):
        """
        n-th order non-central moment of distribution.

        :param n:
        :return:
        """
        return self.scipy_distribution.moment(n, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def nnlf(self, theta, x):
        """
        Return negative loglikelihood function.

        :param theta:
        :param x:
        :return:
        """
        return self.scipy_distribution.nnlf(theta, x)

    @scipy_method
    def ppf(self, q, *args, **kwargs):
        """
        Percent point function(inverse of cdf) at q of the given RV.

        :param q:
        :return:
        """
        return self.scipy_distribution.ppf(q, *args, **kwargs)

    @scipy_method
    def sf(self, x, *args, **kwargs):
        """
        Survival function(1 - cdf) at x of the given RV.

        :param x:
        :return:
        """
        return self.scipy_distribution.sf(x, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def stats(self, *args, **kwargs):
        """
        Some statistics of the given RV.

        :return:
        """
        return self.scipy_distribution.stats(*args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def std(self, *args, **kwargs):
        """
        Standard deviation of the distribution.

        :param args:
        :param kwargs:
        :return:
        """
        return self.scipy_distribution.std(*args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def support(self, *args, **kwargs):
        """
        Return the support of the distribution.

        :param args:
        :param kwargs:
        :return:
        """
        return self.scipy_distribution.support(*args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def var(self, *args, **kwargs):
        """
        Variance of the distribution.

        :return:
        """
        return self.scipy_distribution.var(*args, **kwargs, **self.scipy_distribution_arguments)

    def to_sk_learn(self):
        return self.scipy_distribution


class ScipyContinuousDistributionWrapper(ScipyDistributionWrapper, ContinuousHyperparameterDistribution):
    def __init__(self, scipy_distribution, null_default_value=None, **kwargs):
        ScipyDistributionWrapper.__init__(
            self,
            scipy_distribution=scipy_distribution,
            **kwargs
        )
        ContinuousHyperparameterDistribution.__init__(self, null_default_value=null_default_value)


class ScipyDiscreteDistributionWrapper(ScipyDistributionWrapper, DiscreteHyperparameterDistribution):
    def __init__(self, scipy_distribution, null_default_value=None, **kwargs):
        ScipyDistributionWrapper.__init__(
            self,
            scipy_distribution=scipy_distribution,
            **kwargs
        )
        DiscreteHyperparameterDistribution.__init__(self, null_default_value=null_default_value)

    def values(self):
        if not hasattr(self, "precomp"):
            self.precomp = list(sorted(list(set(
                self.rvs() for _ in range(5000)
            ))))
        return self.precomp


class BaseCustomDiscreteScipyDistribution(ScipyDiscreteDistributionWrapper):
    def __init__(self, name, min_included, max_included, null_default_value, **kwargs):
        scipy_dist_obj = rv_discrete(
            name=name,
            a=min_included,
            b=max_included,
            **kwargs
        )
        scipy_dist_obj._pmf = self._pmf
        ScipyDiscreteDistributionWrapper.__init__(
            self,
            scipy_distribution=scipy_dist_obj,
            null_default_value=null_default_value
        )

    def _override_scipy_methods(self):
        self.scipy_distribution._pmf = self._pmf

    @abstractmethod
    def _pmf(self, x, *args):
        pass


class Distribution(rv_continuous):
    def _pdf(self, x):
        return 0.0


class BaseCustomContinuousScipyDistribution(ScipyContinuousDistributionWrapper):
    def __init__(self, name, min_included, max_included, null_default_value, **kwargs):
        scipy_distribution = Distribution(
            name=name,
            a=min_included,
            b=max_included,
            **kwargs
        )
        scipy_distribution._pdf = self._pdf

        ScipyContinuousDistributionWrapper.__init__(
            self,
            scipy_distribution=scipy_distribution,
            null_default_value=null_default_value
        )

    def _override_scipy_methods(self):
        self.scipy_distribution._pdf = self._pdf

    @abstractmethod
    def _pdf(self, x, *args):
        pass


class ScipyLogUniform(LogSpaceDistributionMixin, BaseCustomContinuousScipyDistribution):
    """
    Get a LogUniform distribution.

    Refer to: :class:`scipy.stats.loguniform`.

    .. seealso::
        :func:`~neuraxle.base.BaseStep.set_hyperparams_space`,
        :class:`ScipyDistributionWrapper`,
    """

    def __init__(self, min_included: float, max_included: float, null_default_value=None):
        """
        Create a quantized random log uniform distribution.
        A random float between the two values inclusively will be returned.

        :param min_included: minimum integer, should be somehow included.
        :param max_included: maximum integer, should be somehow included.
        :param null_default_value: null default value for distribution. if None, take the min_included
        :type null_default_value: int
        """
        if null_default_value is None:
            null_default_value = math.log2(min_included)
        else:
            null_default_value = math.log2(null_default_value)

        self.min_included: float = min_included
        self.max_included: float = max_included
        self.log2_min_included = math.log2(min_included)
        self.log2_max_included = math.log2(max_included)

        super().__init__(
            name='log_uniform',
            null_default_value=null_default_value,
            min_included=self.log2_min_included,
            max_included=self.log2_max_included
        )

    def _pdf(self, x) -> float:
        """
        Calculate the logUniform probability distribution value at position `x`.

        :param x: value where the probability distribution function is evaluated.
        :return: value of the probability distribution function.
        """
        if (self.log2_min_included == self.log2_max_included) and x == 2 ** self.log2_min_included:
            return 1.

        if (x >= 2 ** self.log2_min_included) and (x <= 2 ** self.log2_max_included):
            return 1 / (x * math.log(2) * (self.log2_max_included - self.log2_min_included))

        return 0.


class StdMeanLogNormal(LogSpaceDistributionMixin, BaseCustomContinuousScipyDistribution):
    """
    Get a normal distribution from std and min.

    .. seealso::
        :class:`NormalScipyDistribution`,
        :func:`~neuraxle.base.BaseStep.set_hyperparams_space`,
        :class:`HyperparameterDistribution`,
        :class:`ScipyDistributionWrapper`,
        :class:`neuraxle.hyperparams.space.HyperparameterSamples`,
        :class:`neuraxle.hyperparams.space.HyperparameterSpace`,
        :class:`neuraxle.base.BaseStep`
    """

    def __init__(
            self,
            log2_space_mean: float,
            log2_space_std: float,
            hard_clip_min: float,
            hard_clip_max: float,
            null_default_value: float = None
    ):
        """
        Create a normal distribution from mean and standard deviation.

        :param log2_space_mean: the most common value to pop
        :type log2_space_mean: float
        :param log2_space_std: the standard deviation (that is, the sqrt of the variance).
        :type log2_space_std: float
        :param hard_clip_min: if not none, rvs will return max(result, hard_clip_min).
        :type hard_clip_min: float
        :param hard_clip_max: if not none, rvs will return min(result, hard_clip_min).
        :type hard_clip_max: float
        :param null_default_value: if none, null default value will be set to hard_clip_min.
        :type null_default_value: float
        """
        if null_default_value is None:
            null_default_value = hard_clip_min

        if hard_clip_min is None:
            hard_clip_min = np.nan

        if hard_clip_max is None:
            hard_clip_max = np.nan

        self.log2_space_mean = log2_space_mean
        self.log2_space_std = log2_space_std

        super().__init__(
            name='log_normal',
            min_included=hard_clip_min,
            max_included=hard_clip_max,
            null_default_value=null_default_value
        )

    def _pdf(self, x):
        if x <= 0:
            return 0.

        cdf_min = 0.
        cdf_max = 1.

        pdf_x = 1 / (x * math.log(2) * self.log2_space_std * math.sqrt(2 * math.pi)) * math.exp(
            -(math.log2(x) - self.log2_space_mean) ** 2 / (2 * self.log2_space_std ** 2))
        return pdf_x / (cdf_max - cdf_min)


class Gaussian(BaseCustomContinuousScipyDistribution):
    """
    Gaussian distribution that inherits from `scipy.stats.rv_continuous <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.rv_continuous.html#scipy.stats.rv_continuous>`_

    Example usage :

    .. code-block:: python

        gaussian_distribution = Gaussian(
            min_included=0,
            max_included=10,
            null_default_value=0.0
        )

        assert 0.0 <= gaussian_distribution.rvs() <= 10.0
        assert gaussian_distribution.pdf(10) < 0.001
        assert gaussian_distribution.pdf(0) < 0.42
        assert 0.55 > gaussian_distribution.cdf(5.0) > 0.45
        assert gaussian_distribution.cdf(0) == 0.0


    .. seealso::
        :class:`GaussianScipyDistribution`,
        :class:`ScipyDistributionWrapper`,
        :func:`~neuraxle.base.BaseStep.set_hyperparams_space`,
        :class:`HyperparameterDistribution`,
        :class:`neuraxle.hyperparams.space.HyperparameterSamples`,
        :class:`neuraxle.hyperparams.space.HyperparameterSpace`,
        :class:`neuraxle.base.BaseStep`
    """

    def __init__(self, min_included: int, max_included: int, null_default_value: float = None):
        self.max_included = max_included
        self.min_included = min_included

        BaseCustomContinuousScipyDistribution.__init__(
            self,
            name='gaussian',
            min_included=min_included,
            max_included=max_included,
            null_default_value=null_default_value
        )

    def _pdf(self, x):
        return math.exp(-x ** 2 / 2.) / np.sqrt(2.0 * np.pi)


class Poisson(BaseCustomDiscreteScipyDistribution):
    """
    Poisson distribution that inherits from `scipy.stats.rv_discrete <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.rv_histogram.html#scipy.stats.rv_histogram>`_

    Example usage :

    .. code-block:: python

        poisson_distribution = Poisson(
            min_included=0.0,
            max_included=10.0,
            null_default_value=0.0,
            mu=5.0
        )

        rvs = [poisson_distribution.rvs() for i in range(10)]
        assert not all(x == rvs[0] for x in rvs)
        assert 0.0 <= poisson_distribution.rvs() <= 10.0
        assert poisson_distribution.pdf(10) == 0.01813278870782187
        assert np.isclose(poisson_distribution.pdf(0), 0.006737946999085467)
        assert poisson_distribution.cdf(5.0) == 0.6159606548330632
        assert poisson_distribution.cdf(0) == 0.006737946999085467


    .. seealso::
        :func:`~neuraxle.base.BaseStep.set_hyperparams_space`,
        :class:`PoissonScipyDistribution`,
        :class:`HyperparameterDistribution`,
        :class:`ScipyDistributionWrapper`,
        :class:`~neuraxle.hyperparams.space.HyperparameterSamples`,
        :class:`~neuraxle.hyperparams.space.HyperparameterSpace`,
        :class:`~neuraxle.base.BaseStep`
    """

    def __init__(self, min_included: float, max_included: float, null_default_value: float = None, mu=0.6):
        super().__init__(
            min_included=min_included,
            max_included=max_included,
            name='poisson',
            null_default_value=null_default_value
        )
        self.mu = mu

    def _pmf(self, x):
        return math.exp(-self.mu) * self.mu ** x / factorial(x)


class Histogram(ScipyDiscreteDistributionWrapper):
    """
    Histogram distribution that inherits from `scipy.stats.rv_histogram <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.rv_histogram.html#scipy.stats.rv_histogram>`_

    Example usage :

    .. code-block:: python

        hist_dist = Histogram(
             histogram=np.histogram(data, bins=100),
             null_default_value=0.0
        )

        assert min(data) <= hist_dist.rvs() <= max(data)
        assert 1.0 > hist_dist.pdf(x=1.0) > 0.0
        assert hist_dist.pdf(x=np.max(data)) == 0.0
        assert hist_dist.pdf(x=np.min(data)) < 0.001
        assert hist_dist.cdf(x=np.max(data)) == 1.0
        assert 0.55 > hist_dist.cdf(x=np.median(data)) > 0.45
        assert hist_dist.cdf(x=np.min(data)) == 0.0


    .. seealso::
        :class:`HyperparameterDistribution`,
        :class:`ScipyDistributionWrapper`,
        :class:`Poisson`,
        :class:`Gaussian`,
        :func:`~neuraxle.base.BaseStep.set_hyperparams_space`,
        :class:`~neuraxle.hyperparams.space.HyperparameterSamples`,
        :class:`~neuraxle.hyperparams.space.HyperparameterSpace`,
        :class:`~neuraxle.base.BaseStep`
    """

    def __init__(self, histogram: np.histogram, null_default_value: float = None, **kwargs):
        scipy_distribution = rv_histogram(histogram=histogram, **kwargs)
        ScipyDiscreteDistributionWrapper.__init__(
            self,
            scipy_distribution=scipy_distribution,
            null_default_value=null_default_value
        )



"""
Create Nested Pipelines in Neuraxle
================================================

You can create pipelines within pipelines using the composition design pattern.

This demonstrates how to create pipelines within pipelines, and how to access the steps and their
attributes in the nested pipelines.

For more info, see the `thread here <https://stackoverflow.com/questions/28822756/getting-model-attributes-from-scikit-learn-pipeline/58359509#58359509>`__.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""

import numpy as np
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

from neuraxle.base import Identity
from neuraxle.pipeline import Pipeline


def main():
    np.random.seed(42)
    X = np.random.randint(5, size=(100, 5))

    # Create and fit the pipeline:
    pipeline = Pipeline([
        StandardScaler(),
        Identity(),
        Pipeline([
            Identity(),
            Identity(),  # Note: an Identity step is a step that does nothing.
            Identity(),  # We use it here for demonstration purposes.
            Pipeline([
                Identity(),
                PCA(n_components=2)
            ])
        ])
    ])
    pipeline, X_t = pipeline.fit_transform(X)

    # Get the components:
    pca_components = pipeline["Pipeline"]["Pipeline"][-1].get_wrapped_sklearn_predictor().components_
    assert pca_components.shape == (2, 5)

    # Discussion:
    # https://stackoverflow.com/questions/28822756/getting-model-attributes-from-scikit-learn-pipeline/58359509#58359509


if __name__ == "__main__":
    main()



from neuraxle.pipeline import Pipeline
from neuraxle.steps.loop import ForEach
from neuraxle.steps.misc import TransformCallbackStep, TapeCallbackFunction, FitCallbackStep, \
    FitTransformCallbackStep


def test_fit_for_each_should_fit_all_steps_for_each_data_inputs_expected_outputs():
    tape = TapeCallbackFunction()
    p = Pipeline([
        ForEach(Pipeline([
            FitCallbackStep(tape.callback, ["1"]),
            FitCallbackStep(tape.callback, ["2"]),
        ]))
    ])
    data_inputs = [[0, 1], [1, 2]]
    expected_outputs = [[2, 3], [4, 5]]

    p = p.fit(data_inputs, expected_outputs)

    assert isinstance(p, Pipeline)
    assert tape.get_name_tape() == ["1", "2", "1", "2"]
    assert tape.data == [([0, 1], [2, 3]), ([0, 1], [2, 3]), ([1, 2], [4, 5]), ([1, 2], [4, 5])]


def test_fit_transform_should_fit_transform_all_steps_for_each_data_inputs_expected_outputs():
    tape = TapeCallbackFunction()
    tape_fit = TapeCallbackFunction()
    p = Pipeline([
        ForEach(Pipeline([
            FitTransformCallbackStep(tape.callback, tape_fit, ["1"]),
            FitTransformCallbackStep(tape.callback, tape_fit, ["2"]),
        ]))
    ])
    data_inputs = [[0, 1], [1, 2]]
    expected_outputs = [[2, 3], [4, 5]]

    p, outputs = p.fit_transform(data_inputs, expected_outputs)

    assert tape.get_name_tape() == ["1", "2", "1", "2"]
    assert tape_fit.get_name_tape() == ["1", "2", "1", "2"]
    assert tape_fit.data == [([0, 1], [2, 3]), ([0, 1], [2, 3]), ([1, 2], [4, 5]), ([1, 2], [4, 5])]


def test_transform_should_transform_all_steps_for_each_data_inputs_expected_outputs():
    tape = TapeCallbackFunction()
    p = Pipeline([
        ForEach(Pipeline([
            TransformCallbackStep(tape.callback, ["1"]),
            TransformCallbackStep(tape.callback, ["2"]),
        ]))
    ])
    data_inputs = [[0, 1], [1, 2]]

    outputs = p.transform(data_inputs)

    assert tape.get_name_tape() == ["1", "2", "1", "2"]



"""
Neuraxle's Logging module
====================================
This module contains the Logging class, which is used to log information about the execution of a pipeline.
It is used by the classes inheriting from BaseStep to log information.
It is also modified by the AutoML class and its Trainer and various Trial repositories-related classes
to log info in various folders.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
import logging
import sys
from io import StringIO
from typing import IO, Dict, List, Optional

LOGGER_FORMAT = "[%(asctime)s][%(levelname)-8s][%(name)-8s][%(module)-1s.py:%(lineno)-1d]: %(message)s"
LOGGING_DATETIME_STR_FORMAT = '%Y-%m-%d_%H:%M:%S.%f'
FORMATTER = logging.Formatter(fmt=LOGGER_FORMAT, datefmt=LOGGING_DATETIME_STR_FORMAT)
if sys.version_info.major <= 3 and sys.version_info.minor <= 7:
    logging.basicConfig(format=LOGGER_FORMAT, datefmt=LOGGING_DATETIME_STR_FORMAT, level=logging.INFO)
else:
    logging.basicConfig(format=LOGGER_FORMAT, datefmt=LOGGING_DATETIME_STR_FORMAT, level=logging.INFO, force=True)


NEURAXLE_LOGGER_NAME = "neuraxle"

LOGGER_STRING_IO: Dict[str, StringIO] = {}
LOGGER_FILE_HANDLERS: Dict[str, logging.FileHandler] = {}


class _FilterSTDErr(logging.Filter):
    def filter(self, record):
        return record.levelno < logging.WARN


class NeuraxleLogger(logging.Logger):

    @staticmethod
    def from_identifier(identifier: str) -> 'NeuraxleLogger':
        """
        Returns a logger from an identifier.
        :param identifier: The identifier of the logger
        :return: The logger
        """
        logger: NeuraxleLogger = logging.getLogger(identifier)

        logger.with_string_io(identifier)
        return logger

    def with_string_io(self, identifier: str) -> 'NeuraxleLogger':
        """
        Returns a logger from an identifier.
        :param identifier: The identifier of the logger
        :return: The logger
        """
        if identifier not in LOGGER_STRING_IO:
            string_io = StringIO()
            LOGGER_STRING_IO[identifier] = string_io
            self._add_stream_handler("string_io_handler", string_io)
        return self

    def with_file_handler(self, file_path: str) -> 'NeuraxleLogger':
        """
        Returns a logger from an identifier.
        :param identifier: The identifier of the logger
        :return: The logger
        """
        self.without_file_handler()

        file_handler = logging.FileHandler(file_path)
        file_handler.setLevel('DEBUG')
        self._add_partial_handler(f"file_handler:{file_path}", file_handler, level=logging.DEBUG)

        LOGGER_FILE_HANDLERS[self.name] = file_handler
        return self

    def without_file_handler(self) -> 'NeuraxleLogger':
        """
        Returns a logger from an identifier.
        :param identifier: The identifier of the logger
        :return: The logger
        """
        if self.name in LOGGER_FILE_HANDLERS:
            LOGGER_FILE_HANDLERS[self.name].close()
            self.removeHandler(LOGGER_FILE_HANDLERS[self.name])
            del LOGGER_FILE_HANDLERS[self.name]
        return self

    def read_log_file(self) -> List[str]:
        if self.name not in LOGGER_FILE_HANDLERS:
            raise ValueError(
                f"No file handler for logger named `{self.name}`. Perhaps you forgot to call "
                f"`self.with_file_handler`, or removed the file handler?")
        with open(LOGGER_FILE_HANDLERS[self.name].baseFilename, 'r') as f:
            return f.readlines()

    def with_std_handlers(self) -> 'NeuraxleLogger':
        self._add_stream_handler("errr_handler", sys.stderr, logging.WARN)
        self._add_stream_handler("info_handler", sys.stdout, logging.DEBUG, _FilterSTDErr())
        return self

    def get_scoped_string_history(self) -> str:
        return LOGGER_STRING_IO[self.name].getvalue()

    def get_root_string_history(self) -> str:
        return LOGGER_STRING_IO[NEURAXLE_LOGGER_NAME].getvalue()

    def print_root_string_history(self) -> None:
        print(f"{NEURAXLE_LOGGER_NAME} -> str:\n\n"
              f"{LOGGER_STRING_IO[NEURAXLE_LOGGER_NAME].getvalue()}")

    def _add_stream_handler(
        self,
        handler_name: str,
        stream: IO,
        level: Optional[int] = None,
        _filter: Optional[logging.Filter] = None
    ) -> 'NeuraxleLogger':
        handler = logging.StreamHandler(stream=stream)
        return self._add_partial_handler(handler_name, handler, level, _filter)

    def _add_partial_handler(
        self,
        handler_name: str,
        handler: logging.Handler,
        level: Optional[int] = None,
        _filter: Optional[logging.Filter] = None
    ) -> 'NeuraxleLogger':
        handler.setFormatter(FORMATTER)
        handler.set_name(f"{self.name}.{handler_name}")
        if level is not None:
            handler.setLevel(level)
        if _filter is not None:
            handler.addFilter(_filter)
        self.addHandler(handler)
        return self


logging.setLoggerClass(NeuraxleLogger)

NEURAXLE_ROOT_LOGGER: NeuraxleLogger = NeuraxleLogger.from_identifier(NEURAXLE_LOGGER_NAME)
NEURAXLE_ROOT_LOGGER.setLevel(logging.DEBUG)
NEURAXLE_ROOT_LOGGER.with_std_handlers()



import numpy as np

from neuraxle.base import ExecutionContext as CX
from neuraxle.data_container import DataContainer as DACT
from neuraxle.pipeline import Pipeline
from neuraxle.steps.loop import FlattenForEach
from neuraxle.steps.numpy import MultiplyByN
from neuraxle.steps.output_handlers import OutputTransformerWrapper

DATA_SHAPE = (3, 4)
FLAT_DATA_SHAPE = (3 * 4, )


def test_flatten_for_each_unflatten_should_transform_data_inputs():
    p = FlattenForEach(MultiplyByN(2), then_unflatten=True)
    data_inputs, _ = _create_random_of_shape(DATA_SHAPE)

    outputs = p.transform(data_inputs)

    assert np.array(outputs).shape == DATA_SHAPE
    assert np.array_equal(outputs, data_inputs * 2)


def test_flatten_for_each_should_transform_data_inputs():
    p = FlattenForEach(MultiplyByN(2), then_unflatten=False)
    data_inputs, _ = _create_random_of_shape(DATA_SHAPE)

    outputs = p.transform(data_inputs)

    assert np.array(outputs).shape == FLAT_DATA_SHAPE
    assert np.array_equal(outputs.flatten(), data_inputs.flatten() * 2)


def test_flatten_for_each_should_transform_data_inputs_and_expected_outputs():
    p = FlattenForEach(Pipeline([
        MultiplyByN(2),
        OutputTransformerWrapper(MultiplyByN(3))
    ]))
    # TODO: should use a tape here and ensure that the MultiplyByN received a flat 12 shape only once and not 3*4 things
    data_inputs, expected_outputs = _create_random_of_shape(DATA_SHAPE)

    p, outputs = p.handle_fit_transform(
        DACT(data_inputs=data_inputs, expected_outputs=expected_outputs), CX())

    assert np.array(outputs.data_inputs).shape == DATA_SHAPE
    assert np.array_equal(outputs.data_inputs, data_inputs * 2)
    assert np.array(outputs.expected_outputs).shape == DATA_SHAPE
    assert np.array_equal(outputs.expected_outputs, expected_outputs * 3)


def _create_random_of_shape(shape):
    data_inputs = np.random.random(shape).astype(np.float32)
    expected_outputs = np.random.random(shape).astype(np.float32)
    return data_inputs, expected_outputs
